using System.ComponentModel;

using Blazored.LocalStorage;

using Microsoft.Extensions.Logging;


namespace NSeguin.Dev.Web.Services;

internal sealed class AppDataService : IAppDataService
{
    public AppDataService(ILocalStorageService storage, ILogger<AppDataService> logger)
    {
        Storage = storage;
        Logger = logger;
        Initialized += InitializedEventHandler;
    }

    public event AsyncEventHandler? Initialized;

    private event AsyncEventHandler? AfterInitialized;

    public bool IsInitialized { get; private set; }
    public UserConfiguration? UserConfiguration { get; private set; }
    private ILocalStorageService Storage { get; }
    private ILogger<AppDataService> Logger { get; }

    public ValueTask DoOnceInitializedAsync(AsyncEventHandler action)
    {
        return DoOnceInitializedAsync(() => action.Invoke(this, EventArgs.Empty));
    }

    public ValueTask DoOnceInitializedAsync(Func<ValueTask> action)
    {
        if (IsInitialized)
        {
            Logger.LogInformation("Service is already initialized, running action immediately");
            return action.Invoke();
        }

        AfterInitialized += (_, _) => action.Invoke();
        return ValueTask.CompletedTask;
    }

    public async ValueTask InitializeAsync()
    {
        if (!IsInitialized && Initialized is not null)
        {
            await Initialized.Invoke(this, EventArgs.Empty);
        }

        IsInitialized = true;
        if (AfterInitialized is not null)
        {
            await AfterInitialized.Invoke(this, EventArgs.Empty);
        }
    }

    public ValueTask SaveAsync()
    {
        return SaveUserConfigurationAsync();
    }

    private async ValueTask InitializedEventHandler(object? sender, EventArgs e)
    {
        Logger.LogDebug("Initializing {Type}", nameof(AppDataService));
        await InitializeUserConfigurationAsync();
    }

    private async Task InitializeUserConfigurationAsync()
    {
        Logger.LogInformation("Initializing {Type}", nameof(UserConfiguration));
        try
        {
            UserConfiguration = await GetUserConfigurationAsync();
            if (UserConfiguration is not null)
            {
                Logger.LogDebug("UserConfiguration loaded from local storage");
            }
            else
            {
                UserConfiguration = new UserConfiguration();
                Logger.LogDebug("New UserConfiguration initialized");
            }

            UserConfiguration.PropertyChanged += UserConfigurationPropertyChangedEventHandler;
        }
        catch (Exception exception)
        {
            Logger.LogError(exception, "Failed to initialize {Type}", nameof(UserConfiguration));
        }
    }

    private async void UserConfigurationPropertyChangedEventHandler(
        object? sender,
        PropertyChangedEventArgs e)
    {
        await SaveUserConfigurationAsync();
    }

    private ValueTask<UserConfiguration?> GetUserConfigurationAsync()
    {
        return Storage.GetItemAsync<UserConfiguration?>(Keys.UserConfiguration);
    }

    private ValueTask SaveUserConfigurationAsync()
    {
        Logger.LogDebug("Saving {Type}", nameof(UserConfiguration));
        return Storage.SetItemAsync(Keys.UserConfiguration, UserConfiguration);
    }

    private static class Keys
    {
        public const string UserConfiguration = nameof(UserConfiguration);
    }
}
