using System.Net.Mail;
using System.Threading.Channels;

using Microsoft.AspNetCore.Components;
using Microsoft.AspNetCore.Components.Web;
using Microsoft.Extensions.Options;

using NSeguin.Dev.Web.Components.Shared;

namespace NSeguin.Dev.Web.Services;

internal sealed class ContactService : IContactService, IDisposable
{
    private readonly MailSettings _mailSettings;
    private readonly SmtpClient _smtpClient;
    private readonly Func<HtmlRenderer> _htmlRendererFactory;

    public ContactService(
        IOptions<MailSettings> mailSettings,
        ILogger<ContactService> logger,
        IServiceProvider serviceProvider,
        ILoggerFactory loggerFactory)
    {
        _mailSettings = mailSettings.Value;
        _smtpClient = _mailSettings.Smtp.CreateSmtpClient();
        _htmlRendererFactory = () => new HtmlRenderer(serviceProvider, loggerFactory);
        Logger = logger;
    }

    public MailAddress MailToAddress =>
        new(
            _mailSettings.ToAddress
            ?? throw new ArgumentNullException(nameof(_mailSettings.ToAddress)));

    private ILogger<ContactService> Logger { get; }

    public async Task SubmitContactFormAsync(ContactFormVM contactFormVm)
    {
        Logger.LogInformation(
            "Submitting contact form from {Name} ({Email})",
            contactFormVm.Name,
            contactFormVm.Email);

        var html = await GetHtmlAsync(contactFormVm);
        var mailMessage = new MailMessage
        {
            From = new MailAddress(
                contactFormVm.Email
                ?? throw new ArgumentNullException(nameof(contactFormVm.Email))),
            Subject = $"Contact Form: {contactFormVm.Name}",
            To = {MailToAddress},
            Body = html,
            IsBodyHtml = true
        };

        await SendMailAsync(mailMessage);
    }

    public void Dispose()
    {
        _smtpClient.Dispose();
    }

    private async Task<string> GetHtmlAsync(ContactFormVM contactFormVm)
    {
        await using var htmlRenderer = _htmlRendererFactory();
        var parameters = ParameterView.FromDictionary(
            new Dictionary<string, object?> {{"ContactForm", contactFormVm}});

        var channel = Channel.CreateBounded<string>(
            new BoundedChannelOptions(1) {SingleReader = true, SingleWriter = true});

        await htmlRenderer.Dispatcher.InvokeAsync(
            async () =>
            {
                var output = await htmlRenderer.RenderComponentAsync<ContactFormEmail>(parameters);
                var html = output.ToHtmlString();
                await channel.Writer.WriteAsync(html);
            });

        return await channel.Reader.ReadAsync();
    }

    private async Task SendMailAsync(MailMessage mailMessage)
    {
        Logger.LogDebug(
            "Sending email to {ToAddress} with subject {Subject}",
            mailMessage.To,
            mailMessage.Subject);

        using (mailMessage)
        {
            await _smtpClient.SendMailAsync(mailMessage);
        }
    }
}
